/*************************************************************************************
Bootloader for MegasquirtAVR
9/5/2003 by Dana Scott ( turbo@mustangs.com )
Shamelessly adapted from various sources found at www.avrfreaks.net and also
the atmel butterfly loader.
//-----------------------------------------------------------------------------
The bootloader code is compiled in the normal fashion using the GCC compiler.
That is with the command line of:   make all

At the DOS prompt program the main.hex file into the genboard using the following 
avrdude commands: ( replace the -c parameter with your appropriate cable id )

avrdude -p m128 -c stk200 -t
w efuse 0 0xff
w hfuse 0 0xcc
w lfuse 0 0x2f
quit
avrdude -p m128 -c stk200 -C avrdude.conf -e -f i -i main.hex

Once Avrdude has finished, power down and disconnect the isp cable.
Now power the Genboard up and connect a serial port cable from Genboard to the PC. 
Now run Avrprog.exe on the PC. In the Avrprog window you can select the file that 
you wish to load, either a program file for the flash or data file for the EEPROM. 
The programming session with Avrprog must be ended by clicking on the Exit button 
in the Avrprog window.             

NOTE: When using the avr109 commands (c/C), the marker word will not be written!

**************************************************************************************/
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/io.h>
#include "defines.h" 

//--------------------------------------------
// functions in assembly.S 
extern void page_erase(uint16_t adr);
extern void page_write(uint16_t adr);
extern void reenable_rww(void);
extern void fill_temp_buffer(uint16_t data,uint16_t adr); 
extern uint16_t read_prog_mem_word(uint16_t address);
//--------------------------------------------
// prototypes
static inline uint8_t serial_is_loopback(void);
uint8_t BufferLoad(uint16_t size, uint8_t mem);
void BlockRead(uint16_t size, uint8_t mem);
static inline void signOn(void);
void sendchar(uint8_t data);
uint8_t recchar(void);
//--------------------------------------------
// globals
uint8_t flags;
uint8_t chip_erased;
uint8_t ebuffer[512];       // bigger than we need, who cares, it's free
uint8_t failure,success,busy; // record LED states on non-existant hardware
uint16_t addressB;
uint16_t tempaddressB;
uint16_t addressW;
uint16_t rbindex,size;
//--------------------------------------------
typedef union {     // simplify the bytes to words translation
uint16_t w;
uint8_t b[2];
}W2B;
//--------------------------------------------

// copied from global.h
#define LEDPORT     PORTB
#define LED_STATUS  0 // active LOW!
#define LED_MISC    1 // active HIGH!
#define LED_ACCEL   2
#define LED_WARMUP  4
//#define LED_SQUIRT  5 // next to rs232 connector

#include "ioinit.c"

int main(void)
{
	W2B temp,data;
	uint16_t x;
	uint8_t val;
	//,xx;
	uint8_t ok=1;
	uint16_t ck1, ck2; // protective checksum

    __asm__ __volatile__ ("cli");
    //RAMPZ =1;
    if((read_prog_mem_word((uint16_t)(LAST_APP_BYTE-1))!=(MARKER_WORD)  )  // stay in boot w/ no app loaded
        || (SPCR == 0x35 )                              // stay in boot when commanded by app
        || (SPCR == 0x36 )                              // stay in boot when commanded by app
        || (serial_is_loopback() ) )
	{                   // stay in boot when serial port is looped back
        ;               
    }  
    else 
	{          // else prepare to go to app
        val = MCUCR;              // restore IVSEL to 0. See m128 datasheet pg 60.
        val &= ~(_BV(IVSEL) | _BV(IVCE));  //        "
        flags = val;              //prevents over-optimization by compiler                     "
        MCUCR |= _BV(IVCE);        //        "
        MCUCR = val;              //
        __asm__ ("jmp 0x0000"); //
    }
    
	WDTCR = _BV(WDTOE) | _BV(WDE);
    WDTCR = 0;                  //  disable watchdog 
     
    ACSR &= ( uint8_t)~_BV(ACIE);  //bit 3 low to
    ACSR |=  _BV(ACD);            // bit 7 high to power down the comparitor

    if ( SPCR != 0x36 )
	{
        ioinit();  // set io ports to a safe stat
    
        UBRRH = 0;                                     //baud high
        UBRRL = 51;                                    // 51 for 19200 baud @ 16MHz
                                                    // 47 for 19200 baud @14.7MHz
        UCSRA = 0;
        UCSRB = 0x18; //d8;
        UCSRC = 0x86;                                  // Async. mode, 8N1
    }
    // else {
    // app has configured uart & io
    // }

    LEDPORT |= _BV(LED_MISC); // turn on one external led
    LEDPORT &= ~_BV(LED_STATUS); // turn on the internal led

	GICR = 0;
	//    failure = 1;                                    // init phantom 'LED' states
	//    busy = 1;                                       // init phantom 'LED' states
    for(;;) 
	{
        val = recchar();
        // Check autoincrement status.
		if(val == 'a') 
		{                            //Autoincrement?
            sendchar('Y');                          //Autoincrement is quicker
        }
		 // Set address.
		else if(val == 'A') // Set address...
		{                       //write address
			temp.b[1] = recchar();                  //read address 8 MSB
			temp.b[0] = recchar();                  //read address 8 LSB
			//?RAMPZ = (temp.b[1]&0x80)?1:0;           // set RAMPZ if overflow will occur in word to byte conversion  
			addressW = temp.w;
			addressB = (addressW<<1);               //convert from word address to byte address
			sendchar('\r');
		}
		// Check block load support.
		else if(val == 'b') 
		{                       // Buffer load support
			sendchar('Y');                          // Report buffer load supported
			sendchar((uint8_t)(PAGESIZEBYTES >> 8)) ;
			sendchar((uint8_t)(PAGESIZEBYTES & 0xFF));   // Report buffer size in bytes
		}
		// Start block load.
		else if(val == 'B') 
		{                       // Start buffer load
			temp.b[1] = recchar();                  // Load high byte of buffersize
			temp.b[0] = recchar();                  // Load low byte of buffersize
			ok = recchar();                         // Load memory type ('E' or 'F')
			if (ok == 'F') 
			{                        // 'F' = 0x46
				tempaddressB = addressB;            // copy address 
				do 
				{
					data.b[0] = recchar();
					data.b[1] = recchar();
					fill_temp_buffer(data.w,addressB);
					addressB += 2;                  // Select next word in memory
					temp.w -= 2;                    // Reduce number of bytes to write by two   
				}
				while(temp.w);                     // Loop until all bytes written	
				tempaddressB &= ~(PAGESIZEBYTES-1); // Ensure the address points to the first byte in the page
				page_write(tempaddressB);           // Program page contents
				reenable_rww();                     //Re-enable the RWW section
				if ( addressB & (PAGESIZEBYTES-1))
				{ // Ensure that the address points to the beginning of the next page
					addressB &= ~(PAGESIZEBYTES-1); //  this (0xFF00) changes depending on target device page size in bytes
					addressB += PAGESIZEBYTES;
				}
				//if ( addressB == 0 )                // set rampz to 1 on rollover
			    //?RAMPZ = 1;
				//;
				sendchar('\r');                     // Report programming OK
			}                                       // End FLASH

			// we don't need eeprom programming, the app can do that itself
			if (ok == 'E')	
			{
	            x = temp.w;                         // save size
	            tempaddressB = 0;                   // Start EEPROM
		        do 
				{
		            ebuffer[tempaddressB++]= recchar(); 		
		        }
				while(--temp.w);                   // Loop until all bytes written
	            tempaddressB = 0;                   // Start EEPROM
		        temp.w = x;                         // restore size
		        do 
				{		
	 	            EEARL = addressW;               // Setup EEPROM address
	                EEARH = (addressW >> 8);
					addressW++;                     // Select next byte
					EEDR = ebuffer[tempaddressB++]; // Load data to write
					EECR |= (1<<EEMWE);             // Write data into EEPROM
					EECR |= (1<<EEWE);
	                while (EECR & (1<<EEWE))        // Wait for EEPROM write to finish
	                    ;
		        }
				while(--temp.w);                   // Loop until all bytes written
      	        sendchar('\r');                     // Report programming OK
	        }
		} // do NOT mix with other write-modes! (addressW update issues)
		else if(val == 'D') 
		{		// Start NONSTANDARD ultrafast buffer load
			ok = recchar();			// Load checksum type ('k' expected, or 'l' for verify-only)
			ok |= recchar();		// send 0: reserved for future
			ck1 = ck2 = addressW;	// init checksum
	        tempaddressB = addressB;            // copy address 
			temp.w = 64;			// wired pagesize (words this time)
            do 
			{
                if(ok == 'k')
				{
				    data.b[0] = recchar();
					data.b[1] = recchar();
					fill_temp_buffer(data.w,addressB);
				} 
				else 
				{ //verify flash (calc checksum)
					data.w = read_prog_mem_word(addressB);
				}
                addressB += 2;                  // Select next word in memory
 				ck1 += data.w;		// calculate checksum on the fly
				ck2 ^= data.w;
			    temp.w --;			// how many words left ?
            }
			while(temp.w);		// Loop until all bytes written		
			
			// read (big endian) and check checksum :
			if(ok=='k')
			{
				if(recchar() ^ (ck1 >> 8)) ok = 0; // ^ same as !=
				if(recchar() ^ (ck1 & 0xFF)) ok = 0;
				if(recchar() ^ (ck2 >> 8)) ok = 0;
				if(recchar() ^ (ck2 & 0xFF)) ok = 0;
				if(ok == 0)	// Houston, we got some problem
					sendchar('N');	// report failure
				else 
				{
					sendchar('\r'); // Report programming OK
					tempaddressB &= ~(PAGESIZEBYTES-1); // Ensure the address points to the first byte in the page
					page_write(tempaddressB);           // Program page contents
					reenable_rww();                     //Re-enable the RWW section
				}
			} 
			else 
			{ // sending the calculated checksum
				sendchar(ck1 >> 8);
				sendchar(ck1 & 0xFF);
				sendchar(ck2 >> 8);
				sendchar(ck2 & 0xFF);
			}
			
			addressW += 128;			
	        if ( addressB & (PAGESIZEBYTES-1))
			{ // Ensure that the address points to the beginning of the next page
			    addressB &= ~(PAGESIZEBYTES-1); //  this (0xFF00) changes depending on target device page size in bytes
			    addressB += PAGESIZEBYTES;
				// are we really lost here? What should we do?
		    }
		    //if ( addressB == 0 )                // set rampz to 1 on rollover
			    //?RAMPZ = 1;
			//;
		}
		else if(val == 'g') 
		{                       // Block read	
			temp.b[1] = recchar();                  // Load high byte of size
			temp.b[0] = recchar();                  // Load low byte of size
            val = recchar();                        // Get memtype
            BlockRead(temp.w,val);                  // Perform the block read
        }
        else if(val=='c') 
		{                  //Write program memory, low byte
           data.b[0] = recchar();
           sendchar('\r');
        }

        else if(val== 'C') 
		{                 //Write program memory, high byte
           data.b[1] = recchar();
           fill_temp_buffer(data.w,addressB);
           addressB += 2;
           sendchar('\r');
        }
        else if(val=='e') 
		{                         //Chip erase,  marker word gets erased first
            chip_erased = TRUE;
            addressW = LAST_APP_WORD + PAGESIZEWORDS;// will be marker address
            addressW &= (~(PAGESIZEWORDS-1));       // beginning of marker page 
			do
			{
				addressW -= PAGESIZEWORDS;
                /*if ( addressW & 0x8000 )            // word address
                    //?RAMPZ=1;
                ;
				else
                    //?RAMPZ=0;*/
                page_erase(addressW<<1);            //Perform page erase
                reenable_rww();                     //Re-enable the RWW section
            }
			while(addressW);
            sendchar('\r');
        }
        else if(val=='E') 
		{                         //Exit upgrade
		    sendchar('\r');
		    if (chip_erased == TRUE)
			{ // TODO: find a good condition
			// (busy == 0)&&(failure == 0)&&(success == 1)&&(chip_erased == TRUE)  // check "LED's"
                //?RAMPZ= 1; 
                addressB =(uint16_t)(LAST_APP_BYTE);
                addressB &= (~(PAGESIZEBYTES-1));
                for ( x=0;x<PAGESIZEWORDS;x++) 
				{
                    if ( x == (PAGESIZEWORDS-1))
                        fill_temp_buffer(MARKER_WORD,addressB); // this happens once -> ( last )
                    else
					// MIK: can we be sure that there isn't placed any code in this page?
                        fill_temp_buffer(0xffff,addressB);      // this happens 127 times
                    addressB += 2;                  // Select next word in memory
                }
                //?RAMPZ= 1; 
                addressB =(uint16_t)LAST_APP_BYTE;  // truncates last bit, 
                addressB &= (~(PAGESIZEBYTES-1));
                page_write(addressB);	            // Program page contents, sets the marker word
                reenable_rww();                     //Re-enable the RWW section
                __asm__ ("jmp 0x0000");
            }
            else 
			{
                // if ( chip_erased == FALSE )
                    __asm__ ("jmp 0x0000");
            }
        }
        else if(val== 'm') 
		{                 // write page
           page_write(addressB);           // Program page contents
           reenable_rww();                     //Re-enable the RWW section
           sendchar('\r');
        }
        else if(val=='P' || val=='L') 
		{                         // Enter programming mode
            sendchar('\r');
        }       
        else if (val=='p') 
		{
            sendchar('S');
        }
        else if(val=='t') 
		{                         // Return programmer type
            sendchar(DEVTYPE);
            sendchar(0);
        }
        else if (val=='x')
		{                         // store LED ON states
            ok = recchar();
	    /*            if (ok == 0 ) 
                failure = 1;
            if ( ok == 1 )
                success = 1;
            if ( ok == 2 ) 
	    busy = 1;*/
            sendchar('\r');   
        }
        else if (val=='y') 
		{                        // store LED OFF states
            ok = recchar();
	    /*            if (ok == 0 ) 
                failure = 0;
            if ( ok == 1 )
                success = 0;
            if ( ok == 2 ) 
	    busy = 0;*/
            sendchar('\r');
        }
       	else if (val=='T') 
		{
            recchar();
            sendchar('\r');
       	}
        else if (val=='S') 
		{                        // Return software identifier
            signOn();
        }                    
        else if(val=='R') 
		{                  //Read program memory
            data.w = read_prog_mem_word(addressB);
            sendchar(data.b[1]);                    //send MSB
            sendchar(data.b[0]);                    //send LSB
            addressB += 2;                          // Select next word in memory

            /*if(addressB == 0x0000) 
			{            //higher location it require the use of RAMPZ
                //?RAMPZ++;                       //RAMPZ has to incremented into upper 64k of FLASH memory
                //?RAMPZ &=(1<<RAMPZ0);           //Mask away any non used bits
            ;
			}*/
        }
        else if (val=='v' || val=='V') 
		{	// Return Hardware (SW) Version (uisp likes this!)
            sendchar('2');
            sendchar('0');
        }
        else if (val=='s') 
		{	                    // Return Signature Byte
            sendchar(SIGNATURE_BYTE_1);
            sendchar(SIGNATURE_BYTE_2);
            sendchar(SIGNATURE_BYTE_3);
        }
        else if(val!=0x1b) 
		{                        // if not esc
            sendchar('?');
        }
    }
}


//----------------------------------------------------
uint8_t recchar(void)
{ 
  uint8_t t,again;

  do 
  {
    while(!(UCSRA & (1<<RXC))) // wait for a char to arrive
      ;

    if (UCSRA & _BV(FE)) 
	{
      t = UDR;  // If framing error read UDR
      again = 1;
    } else
      again = 0;

  } while(again);

  return UDR;
}
//----------------------------------------------------
void sendchar(uint8_t data)
{
    while( !(UCSRA&0x20) )        // change back to  0
        ;
    UDR = data; 
}
//----------------------------------------------------
void minWait(void) 
{        // short wait loop for serial_is_loopback() function
  volatile uint8_t x = 40;
      while (--x) 
        ;
}
//----------------------------------------------------
// verify that serial port hardware is looped back.  If you are worried about false positive responses
// you can duplicate pairs of if((PIND...  statements and add as many tests as you like. Maybe even put
// it in a loop if you are really paranoid!
static inline uint8_t serial_is_loopback(void)
{
	uint8_t i;
    UCSRB = 0;                                     // shut off the uart
    DDRD |= _BV(1);                                 // get the directions right
    DDRD &= (uint8_t)~_BV(0);                       // get the directions right

	i = 3;
	do {
		PORTD &= (uint8_t)~_BV(1);                      // set pe1 lo
		minWait();                                      // wait for prop delay
		if (PIND & _BV(0))		// if com rcv hi, we cannot be connected
			return 0;
    
		PORTD |= _BV(1);		// set com send hi
		minWait();
		if ( (PIND & _BV(0)) == 0)	// if com rcv low, we cannot be connected
			return 0;
    
		i--;
	} while(i);

	return 1;
}
//----------------------------------------------------

void BlockRead(uint16_t size, uint8_t mem)
{
W2B data;
	
	if (mem == 'E')	{                               // Read EEPROM
		do {	
			EEARL = addressB;                       // Setup EEPROM address
			EEARH = (addressB >> 8);
			addressB++;                             // Select next EEPROM byte
			EECR |= (uint8_t)(1<<EERE);             // Read EEPROM
			sendchar(EEDR);                         // Transmit EEPROM dat ato PC
			size--;                                 // Decrease number of bytes to read
		}while (size);                              // Repeat until all block has been read
	}
	else  {                                         // Read Flash
        size &= ~_BV(0);                            // wrong (odd) size will get stuck
        while (EECR & 0x02); 
		do {
			data.w = read_prog_mem_word(addressB);
			sendchar(data.b[0]);                    //send LSB
			sendchar(data.b[1]);                    //send MSB  
			addressB += 2;                          // Select next word in memory
			size -= 2;                              // Subtract two bytes from number of bytes to read
		}while (size);                              // Repeat until all block has been read
	}
}
//----------------------------------------------------

static inline void signOn(void)
{ 
    sendchar('A');                                  // Report signon
    sendchar('V');
    sendchar('R');
    sendchar('E');
    sendchar('F');
    sendchar('I');
    sendchar('1');      // I thought that we'd use this digit as a rev level indicator .
                        // change this to a define if you like.
}

/*
MIK: This is handled in assembly.S instead

void bj (void) __attribute__ ((section (".bootjmp")));
// make sure the makefile has this line:
// LDFLAGS += -Wl,--section-start=.bootjmp=0x3ffc
// this way it does not matter what size the bootloader is, it always can be invoked by
// the app when the app does a JMP 0x3ffc
// 0x3ffc is at the end of the mega16 flash, change this value for other processors, is 
// always last_byte-3
// 
void bj(void)   // put the boot jump at the end of memory
{   
    BOOTJMP     // the return instruction needs to get manually trimmed from the hex file
}
*/
